@@ -33,14 +33,11 @@ enum
 	R_COMPLETE,
 	R_BEGINNING_OF_HISTORY,
 	R_END_OF_HISTORY,
-	R_DELETE_LINE,
 	R_BACKWARD_KILL_LINE,
 	R_KILL_WHOLE_LINE,
 	R_KILL_WORD,
 	R_BACKWARD_KILL_WORD,
 	R_DUMP_FUNCTIONS,
-	R_WINCH,
-	R_EXIT,
 	R_HISTORY_TOKEN_SEARCH_BACKWARD,
 	R_HISTORY_TOKEN_SEARCH_FORWARD,
 	R_SELF_INSERT,
@@ -49,7 +46,9 @@ enum
 	R_EXECUTE,
 	R_BEGINNING_OF_BUFFER,
 	R_END_OF_BUFFER,
-	R_REPAINT
+	R_REPAINT,
+	R_UP_LINE,
+	R_DOWN_LINE,
 }
 ;
 
@@ -97,26 +96,42 @@ void input_unreadch( wint_t ch );
    \param d a description of the sequence
    \param cmd an input function that will be run whenever the key sequence occurs
 */
-void add_mapping( const wchar_t *mode, const wchar_t *s, const wchar_t * d, const wchar_t *cmd );
+void input_mapping_add( const wchar_t *sequence, const wchar_t *cmd );
+
+void input_mapping_get_names( array_list_t *list );
+
+int input_mapping_erase( const wchar_t *sequence );
+
+const wchar_t *input_mapping_get( const wchar_t *sequence );
 
 /**
-   Sets the mode keybindings. 
-*/
-void input_set_mode( wchar_t *name );
+   Return the sequence for the terminfo variable of the specified name.
+
+   If no terminfo variable of the specified name could be found, return 0 and set errno to ENOENT.
+   If the terminfo variable does not have a value, return 0 and set errno to EILSEQ.
+ */
+const wchar_t *input_terminfo_get_sequence( const wchar_t *name );
 
 /**
-   Sets the application keybindings
-*/
-void input_set_application( wchar_t *name );
+   Return the name of the terminfo variable with the specified sequence
+ */
+const wchar_t *input_terminfo_get_name( const wchar_t *seq );
 
 /**
-   Parse a single line of inputrc information. 
-*/
-void input_parse_inputrc_line( wchar_t *cmd );
+   Return a list of all known terminfo names
+ */
+void input_terminfo_get_names( array_list_t *lst, int skip_null );
+
 
 /**
-   Returns the function for the given function name.
+   Returns the input function code for the given input function name.
 */
-wchar_t input_get_code( const wchar_t *name );
+wchar_t input_function_get_code( const wchar_t *name );
+
+/**
+   Returns a list of all existing input function names
+ */
+void input_function_get_names( array_list_t *lst );
+
 
 #endif
