@@ -93,6 +93,23 @@ void parse_util_token_extent( const wchar_t *buff,
 */
 int parse_util_lineno( const wchar_t *str, int len );
 
+/**
+   Calculate the line number of the specified cursor position
+ */
+int parse_util_get_line_from_offset( wchar_t *buff, int pos );
+
+/**
+   Get the offset of the first character on the specified line
+ */
+int parse_util_get_offset_from_line( wchar_t *buff, int line );
+
+
+/**
+   Return the total offset of the buffer for the cursor position nearest to the specified poition
+ */
+int parse_util_get_offset( wchar_t *buff, int line, int line_offset );
+
+
 /**
    Autoload the specified file, if it exists in the specified path. Do
    not load it multiple times unless it's timestamp changes or
@@ -138,7 +155,7 @@ int parse_util_unload( const wchar_t *cmd,
    Set the argv environment variable to the specified null-terminated
    array of strings. 
 */
-void parse_util_set_argv( wchar_t **argv );
+void parse_util_set_argv( wchar_t **argv, array_list_t *named_arguments );
 
 /**
    Make a duplicate of the specified string, unescape wildcard
@@ -146,4 +163,6 @@ void parse_util_set_argv( wchar_t **argv );
 */
 wchar_t *parse_util_unescape_wildcards( const wchar_t *in );
 
+
+
 #endif
